/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module_Atan.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>
#include <math.h>


// define some constants to make referencing in/outs clearer
#define PN_INPUT1 0
#define PN_OUTPUT1 1

// The most important part, processing the audio
void Module_Atan::sub_process(long buffer_offset, long sampleFrames )
{
	// assign some pointers to your in/output buffers. usually blocks (array) of 96 samples
    float *in1  = buffer_offset + input1_buffer;
    float *out1  = buffer_offset + output1_buffer;

    while(--sampleFrames >= 0) // sampleFrames = how many samples to process (can vary). repeat (loop) that many times
    {
    *out1 = (float)atan(*in1 * 10.0f) / 10.0f;

    // increment the pointers (move to next sample in buffers)
    in1++;
    out1++;
    }
}

void Module_Atan::sub_process_static(long buffer_offset, long sampleFrames )
{
	sub_process(buffer_offset, sampleFrames);
	static_count = static_count - sampleFrames;
	if( static_count <= 0 /* && !AreEventsPending()*/ )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

//Module::Module(seaudioMasterCallback seaudioMaster) : SEModule_base(seaudioMaster)
Module_Atan::Module_Atan(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}

void Module_Atan::open()
{
	SEModule_base::open();

	// choose which function is used to process audio
	SET_PROCESS_FUNC(Module_Atan::sub_process);
}

void Module_Atan::close()
{
}

Module_Atan::~Module_Atan()
{
}

// describe your module
bool Module_Atan::getModuleProperties (SEModuleProperties* properties)
{
	// describe the plugin, this is the name the end-user will see.
	properties->name = "EZMath Atan";

	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug.
	// No two plugs may have the same id.
	properties->id = "BAVII EZMath Atan";

	// Info, may include Author, Web page whatever
#if defined(MINGW) || defined(__GNUC__)
	properties->about = "by Brian Allen Vanderburg II (GNU)";								// !!TODO!!
#elif defined _MSC_VER //hmm don't seem to work
	properties->about = "by Brian Allen Vanderburg II (MS)";
#else
	properties->about = "by  Brian Allen Vanderburg II (Borland?)";
#endif
	return true;
}

// describe the pins (plugs and parameters)
bool Module_Atan::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )								// !!TODO!! list your in / out plugs
	{
		// typical input plug (inputs are listed first)
	case PN_INPUT1:
		properties->name				= "in";
		properties->variable_address	= &input1_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
		// typical output plug
	case PN_OUTPUT1:
		properties->name				= "out";
		properties->variable_address	= &output1_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

// An input plug has changed value
void Module_Atan::OnPlugStateChange(SEPin *pin)
{
	state_type in_stat1 = getPin(PN_INPUT1)->getStatus();

	// we need to pass on the state of this module's output signal
	// it depends on the inputs...

    // 'transmit' this modules new output status to next module 'downstream'
	getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), in_stat1);

	// setup up 'sleep mode' or not
	if( in_stat1 < ST_RUN )
	{
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module_Atan::sub_process_static);
	}
	else
	{
		SET_PROCESS_FUNC(Module_Atan::sub_process);
	}
}
